package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MixMaterialRepository
import dev.fyloz.colorrecipesexplorer.service.utils.findDuplicated
import dev.fyloz.colorrecipesexplorer.service.utils.hasGaps
import org.springframework.context.annotation.Lazy
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service

interface MixMaterialService : ModelService<MixMaterial, MixMaterialRepository> {
    /** Checks if one or more mix materials have the given [material]. */
    fun existsByMaterial(material: Material): Boolean

    /** Creates [MixMaterial]s from the givens [MixMaterialDto]. */
    fun create(mixMaterials: Set<MixMaterialDto>): Set<MixMaterial>

    /** Creates a [MixMaterial] from a given [MixMaterialDto]. */
    fun create(mixMaterial: MixMaterialDto): MixMaterial

    /** Updates the [quantity] of the given [mixMaterial]. */
    fun updateQuantity(mixMaterial: MixMaterial, quantity: Float): MixMaterial

    /**
     *  Validates if the given [mixMaterials]. To be valid, the position of each mix material must be greater or equals to 1 and unique in the set.
     *  There must also be no gap between the positions. Also, the quantity of the first mix material in the set must not be expressed in percentages.
     *  If any of those criteria are not met, an [InvalidGroupStepsPositionsException] will be thrown.
     */
    fun validateMixMaterials(mixMaterials: Set<MixMaterial>)
}

@Service
class MixMaterialServiceImpl(
  mixMaterialRepository: MixMaterialRepository,
  @Lazy val materialService: MaterialService
) : AbstractModelService<MixMaterial, MixMaterialRepository>(mixMaterialRepository), MixMaterialService {
    override fun idNotFoundException(id: Long) = mixMaterialIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = mixMaterialIdAlreadyExistsException(id)

    override fun existsByMaterial(material: Material): Boolean = repository.existsByMaterial(material)
    override fun create(mixMaterials: Set<MixMaterialDto>): Set<MixMaterial> =
      mixMaterials.map(::create).toSet()

    override fun create(mixMaterial: MixMaterialDto): MixMaterial =
      mixMaterial(
        material = materialService.getById(mixMaterial.materialId),
        quantity = mixMaterial.quantity,
        position = mixMaterial.position
      )

    override fun updateQuantity(mixMaterial: MixMaterial, quantity: Float) =
      update(mixMaterial.apply {
          this.quantity = quantity
      })

    override fun validateMixMaterials(mixMaterials: Set<MixMaterial>) {
        if (mixMaterials.isEmpty()) return

        val sortedMixMaterials = mixMaterials.sortedBy { it.position }
        val firstMixMaterial = sortedMixMaterials[0]
        val errors = mutableSetOf<InvalidMixMaterialsPositionsError>()

        // Check if the first mix material position is 1
        fun isFirstMixMaterialPositionInvalid() =
          sortedMixMaterials[0].position != 1

        // Check if the first mix material is expressed in percents
        fun isFirstMixMaterialPercentages() =
          sortedMixMaterials[0].material.materialType!!.usePercentages

        // Check if any positions is duplicated
        fun getDuplicatedPositionsErrors() =
          sortedMixMaterials
            .findDuplicated { it.position }
            .map { duplicatedMixMaterialsPositions(it) }

        // Find all errors and throw if there is any
        if (isFirstMixMaterialPositionInvalid()) errors += invalidFirstMixMaterialPosition(sortedMixMaterials[0])
        errors += getDuplicatedPositionsErrors()
        if (errors.isEmpty() && mixMaterials.hasGaps { it.position }) errors += gapBetweenStepsPositions()
        if (errors.isNotEmpty()) {
            throw InvalidMixMaterialsPositionsException(errors)
        }

        if (isFirstMixMaterialPercentages()) {
            throw InvalidFirstMixMaterial(firstMixMaterial)
        }
    }
}

class InvalidMixMaterialsPositionsError(
  val type: String,
  val details: String
)

class InvalidMixMaterialsPositionsException(
  val errors: Set<InvalidMixMaterialsPositionsError>
) : RestException(
  "invalid-mixmaterial-position",
  "Invalid mix materials positions",
  HttpStatus.BAD_REQUEST,
  "The position of mix materials are invalid",
  mapOf(
    "invalidMixMaterials" to errors
  )
)

class InvalidFirstMixMaterial(
  val mixMaterial: MixMaterial
) : RestException(
  "invalid-mixmaterial-first",
  "Invalid first mix material",
  HttpStatus.BAD_REQUEST,
  "The first mix material is invalid because its material must not be expressed in percents",
  mapOf(
    "mixMaterial" to mixMaterial
  )
)

const val INVALID_FIRST_MIX_MATERIAL_POSITION_ERROR_CODE = "first"
const val DUPLICATED_MIX_MATERIALS_POSITIONS_ERROR_CODE = "duplicated"
const val GAP_BETWEEN_MIX_MATERIALS_POSITIONS_ERROR_CODE = "gap"

private fun invalidFirstMixMaterialPosition(mixMaterial: MixMaterial) =
  InvalidMixMaterialsPositionsError(
    INVALID_FIRST_MIX_MATERIAL_POSITION_ERROR_CODE,
    "The position ${mixMaterial.position} is under the minimum of 1"
  )

private fun duplicatedMixMaterialsPositions(position: Int) =
  InvalidMixMaterialsPositionsError(
    DUPLICATED_MIX_MATERIALS_POSITIONS_ERROR_CODE,
    "The position $position is duplicated"
  )

private fun gapBetweenStepsPositions() =
  InvalidMixMaterialsPositionsError(
    GAP_BETWEEN_MIX_MATERIALS_POSITIONS_ERROR_CODE,
    "There is a gap between mix materials positions"
  )

