package dev.fyloz.colorrecipesexplorer.service.files

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.slf4j.Logger
import org.springframework.core.io.Resource
import org.springframework.core.io.ResourceLoader
import org.springframework.web.multipart.MultipartFile
import java.io.File
import java.io.IOException
import java.io.InputStream
import java.nio.file.Paths
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class FileServiceTest {
    private val resourcesLoader = mock<ResourceLoader>()
    private val logger = mock<Logger>()
    private val properties = CreProperties()

    private val service = spy(FileService(resourcesLoader, properties, logger))

    private val path = "/var/cre/file"

    @AfterEach
    fun afterEach() {
        reset(resourcesLoader, logger, service)
    }

    // readResource()

    @Test
    fun `readResource() returns content of the resource at the given path`() {
        val resource = mock<Resource>()
        val resourceStream = mock<InputStream>()
        val resourceContent = """
            Line 1
            Line 2
            Line 3
        """.trimIndent()

        whenever(resource.inputStream).doReturn(resourceStream)
        whenever(resourcesLoader.getResource("classpath:$path")).doReturn(resource)
        doReturn(resourceContent).whenever(service).readInputStreamAsString(resourceStream)

        val found = service.readResource(path)

        assertEquals(resourceContent, found)
    }

    // readInputStreamAsString()

    @Test
    fun `readInputStreamAsString() returns a String matching the given input stream's content`() {
        val stream = mock<InputStream>()
        val streamContent = """
            Line 1
            Line 2
            Line 3
        """.trimIndent()

        whenever(stream.readAllBytes()).doAnswer { streamContent.toByteArray() }

        val found = service.readInputStreamAsString(stream)

        assertEquals(streamContent, found)
    }

    // write()

    private inline fun withMultipartFile(size: Long = 1000L, test: (MultipartFile) -> Unit) {
        val multipartFile = mock<MultipartFile>()
        whenever(multipartFile.size).doReturn(size)

        test(multipartFile)
    }

    @Test
    fun `write() transfers data from the given MultipartFile to the file at the given path and returns true`() {
        withMultipartFile { multipartFile ->
            val file = mock<File>()
            val filePath = Paths.get(path)

            whenever(file.toPath()).doReturn(filePath)
            doAnswer { file }.whenever(service).create(path)

            assertTrue { service.write(multipartFile, path) }

            verify(multipartFile).transferTo(filePath)
        }
    }

    @Test
    fun `write() returns true when given MultipartFile is empty`() {
        withMultipartFile(size = 0L) { multipartFile ->
            assertTrue { service.write(multipartFile, path) }

            verify(multipartFile, never()).transferTo(any<File>())
        }
    }

    @Test
    fun `write() returns false when the data transfer throw an IOException`() {
        withMultipartFile { multipartFile ->
            val file = mock<File>()
            val filePath = Paths.get(path)

            whenever(file.toPath()).doReturn(filePath)
            whenever(multipartFile.transferTo(filePath)).doThrow(IOException())
            doAnswer { file }.whenever(service).create(path)

            assertFalse { service.write(multipartFile, path) }
        }
    }
}
