package dev.fyloz.colorrecipesexplorer.service.files

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.model.material
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.web.multipart.MultipartFile
import java.io.IOException
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class SimdutServiceTest {
    private val fileService = mock<FileService>()
    private val service = spy(SimdutService(fileService, mock()))

    private val material = material(id = 0L)

    @AfterEach
    fun afterEach() {
        reset(fileService, service)
    }

    @JvmName("withNullableMaterialPath")
    private inline fun withMaterialPath(material: Material? = null, exists: Boolean = true, test: (String) -> Unit) =
      withMaterialPath(material ?: this.material, exists, test)

    private inline fun withMaterialPath(material: Material, exists: Boolean = true, test: (String) -> Unit) {
        val path = "data/simdut/${material.id}"
        doReturn(path).whenever(service).getPath(material)
        whenever(fileService.exists(path)).doReturn(exists)

        test(path)
    }

    // exists()

    @Test
    fun `exists() returns true when a SIMDUT file exists for the given material`() {
        withMaterialPath {
            assertTrue { service.exists(material) }
        }
    }

    @Test
    fun `exists() returns false when no SIMDUT file exists for the given material`() {
        withMaterialPath(exists = false) {
            assertFalse { service.exists(material) }
        }
    }

    // read()

    @Test
    fun `read() returns a filled ByteArray when a SIMDUT exists for the given material`() {
        withMaterialPath { path ->
            val simdutContent = byteArrayOf(0xf)

            whenever(fileService.readAsBytes(path)).doReturn(simdutContent)

            val found = service.read(material)

            assertEquals(simdutContent, found)
        }
    }

    @Test
    fun `read() returns a empty ByteArray when no SIMDUT exists for the given material`() {
        withMaterialPath(exists = false) {
            val found = service.read(material)

            assertTrue { found.isEmpty() }
        }
    }

    @Test
    fun `read() returns a empty ByteArray when reading the SIMDUT throws an IOException`() {
        withMaterialPath { path ->
            whenever(fileService.readAsBytes(path)).doAnswer { throw IOException() }

            val found = service.read(material)

            assertTrue { found.isEmpty() }
        }
    }

    // write()

    @Test
    fun `write() writes the given MultipartFile to the disk for the given material`() {
        withMaterialPath { path ->
            val simdutMultipart = mock<MultipartFile>()

            whenever(fileService.write(simdutMultipart, path)).doReturn(true)

            service.write(material, simdutMultipart)

            verify(fileService).write(simdutMultipart, path)
        }
    }

    @Test
    fun `write() throws a SimdutWriteException when writing the given MultipartFile to the disk fails`() {
        withMaterialPath { path ->
            val simdutMultipart = mock<MultipartFile>()

            whenever(fileService.write(simdutMultipart, path)).doReturn(false)

            assertThrows<SimdutWriteException> { service.write(material, simdutMultipart) }
        }
    }

    // update()

    @Test
    fun `update() deletes and write the SIMDUT for the given material`() {
        val simdutMultipart = mock<MultipartFile>()

        // Prevents calling the actual implementation
        doAnswer { }.whenever(service).delete(material)
        doAnswer { }.whenever(service).write(material, simdutMultipart)

        service.update(simdutMultipart, material)

        verify(service).delete(material)
        verify(service).write(material, simdutMultipart)
    }

    // delete()

    @Test
    fun `delete() deletes the SIMDUT of the given material from the disk`() {
        withMaterialPath { path ->
            service.delete(material)

            verify(fileService).delete(path)
        }
    }

    // getPath()

    @Test
    fun `getPath() returns the appropriate path for the given material`() {
        val simdutFileName = material.id.toString()
        val workingDirectory = "data"
        val expectedPath = "$workingDirectory/$SIMDUT_DIRECTORY/$simdutFileName"

        whenever(fileService.getPath(any())).doAnswer { "$workingDirectory/${it.arguments[0]}" }
        doAnswer { simdutFileName }.whenever(service).getSimdutFileName(material)

        val found = service.getPath(material)

        assertEquals(expectedPath, found)

        verify(fileService).getPath("$SIMDUT_DIRECTORY/$simdutFileName")
    }
}
